import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], "../../"))
import rips


def test_well_path(rips_instance, initialize_test):
    well_path_coll = rips_instance.project.descendants(rips.WellPathCollection)[0]
    assert len(well_path_coll.well_paths()) == 0

    well_path = well_path_coll.add_new_object(rips.ModeledWellPath)
    well_path2 = well_path_coll.add_new_object(rips.ModeledWellPath)
    assert len(well_path_coll.well_paths()) == 2

    well_path.delete()
    assert len(well_path_coll.well_paths()) == 1

    try:
        # Delete again should throw exception
        well_path.delete()
        assert False
    except Exception:
        assert True

    well_path2.delete()
    assert len(well_path_coll.well_paths()) == 0



import os
import rips

resinsight = rips.Instance.find()

# ResInsight includes some test models. Adjust this path to fit your system
test_model_path = "e:/gitroot-second/ResInsight/TestModels"

case_paths = []
case_paths.append(test_model_path + "/Case_with_10_timesteps/Real0/BRUGGE_0000.EGRID")
case_paths.append(test_model_path + "/Case_with_10_timesteps/Real10/BRUGGE_0010.EGRID")
case_paths.append(test_model_path + "/Case_with_10_timesteps/Real30/BRUGGE_0030.EGRID")
case_paths.append(test_model_path + "/Case_with_10_timesteps/Real40/BRUGGE_0040.EGRID")

for case_path in case_paths:
    assert os.path.exists(
        case_path
    ), "You need to set valid case paths for this script to work"

case_group = resinsight.project.create_grid_case_group(case_paths=case_paths)

cases = case_group.descendants(rips.EclipseCase)
print("Got " + str(len(cases)) + " cases: ")

for case in cases:
    time_step_info = case.time_steps()
    porv_results = case.active_cell_property("STATIC_NATIVE", "PORV", 0)

    for time_step_index in range(0, len(time_step_info)):
        pressure_results = case.active_cell_property(
            "DYNAMIC_NATIVE", "PRESSURE", time_step_index
        )

        results = []
        for pressure, porv in zip(pressure_results, porv_results):
            results.append(pressure * porv)

        # set the computed values in the case
        case.set_active_cell_property(
            results, "GENERATED", "PRESSURE_PORV", time_step_index
        )

    print(
        "Case id: " + str(case.id),
        "  Case name: " + case.name,
        " : Calculation complete",
    )


print("Transferred all results back to ResInsight")

# one of "GENERATED", "DYNAMIC_NATIVE", "STATIC_NATIVE", "IMPORTED"
# https://api.resinsight.org/en/main/rips.html#result-definition
property_type = "GENERATED"

property_name = "PRESSURE_PORV"

statistics_case = case_group.create_statistics_case()
statistics_case.set_source_properties(property_type, [property_name])
statistics_case.compute_statistics()

view = statistics_case.create_view()
statistics_property_name = property_name + "_MEAN"
view.apply_cell_result(
    result_type=property_type, result_variable=statistics_property_name
)



import sys
import os
import tempfile

sys.path.insert(1, os.path.join(sys.path[0], "../../"))
import rips

import dataroot


def test_10k(rips_instance, initialize_test):
    case_root_path = dataroot.PATH + "/TEST10K_FLT_LGR_NNC"
    project_path = case_root_path + "/well_completions_pytest.rsp"
    project = rips_instance.project.open(path=project_path)

    export_folder = tempfile.gettempdir()

    rips_instance.set_export_folder(export_type="COMPLETIONS", path=export_folder)

    case = project.cases()[0]
    case.export_well_path_completions(
        time_step=1,
        well_path_names=["Well-1"],
        file_split="UNIFIED_FILE",
    )


def test_add_well_path_completions(rips_instance, initialize_test):
    well_path_coll = rips_instance.project.descendants(rips.WellPathCollection)[0]

    well_path = well_path_coll.add_new_object(rips.ModeledWellPath)
    well_path.name = "test"
    well_path.update()

    # Update the completion settings
    completions_settings = well_path.completion_settings()
    completions_settings.allow_well_cross_flow = True
    completions_settings.auto_well_shut_in = "STOP"
    completions_settings.drainage_radius_for_pi = 1.56
    completions_settings.fluid_in_place_region = 99
    completions_settings.gas_inflow_eq = "R-G"
    completions_settings.group_name_for_export = "TestGroup"
    completions_settings.hydrostatic_density = "AVG"
    completions_settings.msw_liner_diameter = 0.12
    completions_settings.msw_roughness = 4.66
    completions_settings.reference_depth_for_export = 1234
    completions_settings.well_bore_fluid_pvt_table = 33
    completions_settings.well_name_for_export = "TestWellName"
    completions_settings.well_type_for_export = "LIQUID"
    completions_settings.update()  # Commit updates back to ResInsight

    completions_settings_updated = well_path.completion_settings()
    assert completions_settings_updated.allow_well_cross_flow == True
    assert completions_settings_updated.auto_well_shut_in == "STOP"
    assert completions_settings_updated.drainage_radius_for_pi == "1.56"
    assert completions_settings_updated.fluid_in_place_region == 99
    assert completions_settings_updated.gas_inflow_eq == "R-G"
    assert completions_settings_updated.group_name_for_export == "TestGroup"
    assert completions_settings_updated.hydrostatic_density == "AVG"
    assert completions_settings_updated.msw_liner_diameter == 0.12
    assert completions_settings_updated.msw_roughness == 4.66
    assert completions_settings_updated.reference_depth_for_export == "1234"
    assert completions_settings_updated.well_bore_fluid_pvt_table == 33
    assert completions_settings_updated.well_name_for_export == "TestWellName"
    assert completions_settings_updated.well_type_for_export == "LIQUID"

    msw_settings = well_path.msw_settings()
    msw_settings.custom_values_for_lateral = True
    msw_settings.enforce_max_segment_length = True
    msw_settings.liner_diameter = 20.0
    msw_settings.max_segment_length = 123.05
    msw_settings.pressure_drop = "HFA"
    msw_settings.reference_md_type = "UserDefined"
    msw_settings.roughness_factor = 1.3
    msw_settings.user_defined_reference_md = 1234.56
    msw_settings.update()

    msw_settings_updated = well_path.msw_settings()
    assert msw_settings_updated.custom_values_for_lateral == True
    assert msw_settings_updated.enforce_max_segment_length == True
    assert msw_settings_updated.liner_diameter == 20.0
    assert msw_settings_updated.max_segment_length == 123.05
    assert msw_settings_updated.pressure_drop == "HFA"
    assert msw_settings_updated.reference_md_type == "UserDefined"
    assert msw_settings_updated.roughness_factor == 1.3
    assert msw_settings_updated.user_defined_reference_md == 1234.56


def test_add_well_path_fracture_template(rips_instance, initialize_test):

    # Add test for all properties
    # Some properties depend on availablility of other data and is not tested, these tests are commented out

    fracture_template = rips_instance.project.descendants(rips.FractureTemplate)[0]
    fracture_template.azimuth_angle = 23.0
    # fracture_template.beta_factor_type = "FractureBetaFactor"
    fracture_template.conductivity_factor = 12.5
    fracture_template.conductivity_type = "FiniteConductivity"
    fracture_template.d_factor_scale_factor = 1.2
    fracture_template.effective_permeability = 55
    fracture_template.fracture_width = 0.5
    fracture_template.fracture_width_type = "UserDefinedWidth"
    fracture_template.gas_viscosity = 0.1
    fracture_template.height_scale_factor = 1.2
    fracture_template.height_scale_factor = 4
    fracture_template.inertial_coefficient = 0.7
    fracture_template.non_darcy_flow_type = "Computed"
    fracture_template.orientation = "Azimuth"
    fracture_template.perforation_length = 5
    fracture_template.permeability_type = "UserDefinedPermeability"
    fracture_template.relative_gas_density = 0.1
    fracture_template.relative_permeability = 0.2
    fracture_template.user_defined_d_factor = 14
    fracture_template.user_defined_perforation_length = True
    fracture_template.user_description = "my frac name"
    fracture_template.width_scale_factor = 7

    fracture_template.update()

    fracture_template_updated = rips_instance.project.descendants(
        rips.FractureTemplate
    )[0]
    assert fracture_template_updated.azimuth_angle == 23.0
    # assert fracture_template_updated.beta_factor_type == "FractureBetaFactor"
    assert fracture_template_updated.conductivity_factor == 12.5
    assert fracture_template_updated.conductivity_type == "FiniteConductivity"
    assert fracture_template_updated.d_factor_scale_factor == 1.2
    assert fracture_template_updated.effective_permeability == 55
    assert fracture_template_updated.fracture_width == 0.5
    assert fracture_template_updated.fracture_width_type == "UserDefinedWidth"
    assert fracture_template_updated.gas_viscosity == 0.1
    assert fracture_template_updated.height_scale_factor == 4
    assert fracture_template_updated.inertial_coefficient == 0.7
    assert fracture_template_updated.non_darcy_flow_type == "Computed"
    assert fracture_template_updated.orientation == "Azimuth"
    assert fracture_template_updated.perforation_length == 5
    assert fracture_template_updated.permeability_type == "UserDefinedPermeability"
    assert fracture_template_updated.relative_gas_density == 0.1
    assert fracture_template_updated.relative_permeability == 0.2
    assert fracture_template_updated.user_defined_d_factor == 14
    assert fracture_template_updated.user_defined_perforation_length == True
    assert fracture_template_updated.user_description == "my frac name"
    assert fracture_template_updated.width_scale_factor == 7


# test fishbone interface
def test_fishbone_interface(rips_instance, initialize_test):
    well_path_coll = rips_instance.project.descendants(rips.WellPathCollection)[0]

    well_path = well_path_coll.add_new_object(rips.ModeledWellPath)
    well_path.name = "fishbone_well"
    well_path.update()

    fishbones_collection = well_path.completions().fishbones()
    fishbones_collection.main_bore_skin_factor = 0.1
    fishbones_collection.update()

    fishbones_collection.set_fixed_start_location(2900.0)

    sub_locations2 = [3500.0, 3550.0, 3600.0, 3700.0]
    fishbones = fishbones_collection.append_fishbones(sub_locations2)

    fishbones.lateral_diameter = 21.0
    fishbones.lateral_skin_factor = 2.1
    fishbones.icd_count = 1
    fishbones.icd_flow_coefficient = 0.1
    fishbones.icd_orifice_diameter = 0.1
    fishbones.lateral_build_angle = 0.1
    fishbones.lateral_count_per_sub = 1
    fishbones.lateral_exit_angle = 0.1
    fishbones.lateral_install_success_fraction = 0.1
    fishbones.lateral_length = "1"
    fishbones.lateral_open_hole_roghness_factor = 0.2
    fishbones.lateral_tubing_diameter = 0.1
    fishbones.lateral_tubing_roghness_factor = 0.1
    fishbones.subs_orientation_mode = "FIXED"
    fishbones.update()

    fishbones_updated = fishbones_collection.fishbones()[0]
    assert fishbones_updated.lateral_diameter == 21.0
    assert fishbones_updated.lateral_skin_factor == 2.1
    assert fishbones_updated.icd_count == 1
    assert fishbones_updated.icd_flow_coefficient == 0.1
    assert fishbones_updated.icd_orifice_diameter == 0.1
    assert fishbones_updated.lateral_build_angle == 0.1
    assert fishbones_updated.lateral_count_per_sub == 1
    assert fishbones_updated.lateral_exit_angle == 0.1
    assert fishbones_updated.lateral_install_success_fraction == 0.1
    assert fishbones_updated.lateral_length == "1"
    assert fishbones_updated.lateral_open_hole_roghness_factor == 0.2
    assert fishbones_updated.lateral_tubing_diameter == 0.1
    assert fishbones_updated.lateral_tubing_roghness_factor == 0.1
    assert fishbones_updated.subs_orientation_mode == "FIXED"



# Load ResInsight Processing Server Client Library
import rips

# Connect to ResInsight instance
resinsight = rips.Instance.find()

# Create a modeled well path and add well path targets
# The coordinates are based on the Norne case

well_path_coll = resinsight.project.descendants(rips.WellPathCollection)[0]
well_path = well_path_coll.add_new_object(rips.ModeledWellPath)
well_path.name = "Test Well-1"
well_path.update()

geometry = well_path.well_path_geometry()

reference_point = geometry.reference_point
reference_point[0] = 457196
reference_point[1] = 7322270
reference_point[2] = 2742
geometry.update()  # Commit updates back to ResInsight

# Create the first well target at the reference point
coord = [0, 0, 0]
geometry.append_well_target(coord)

# Append new well targets relative the the reference point
coord = [454.28, 250, -10]
target = geometry.append_well_target(coord)

coord = [1054.28, 250, -50]
target = geometry.append_well_target(coord)

well_path.append_perforation_interval(3300, 3350, 0.2, 0.76)


# Update skin factor of the perforation
perforation_coll = well_path.completions().perforations()
perforation = perforation_coll.perforations()[0]
new_skin_factor = 0.9
print(
    "Changing perforation skin factor from {} to {}.".format(
        perforation.skin_factor, new_skin_factor
    )
)
perforation.skin_factor = new_skin_factor
perforation.update()

# Optionally update the completion settings
completions_settings = well_path.completion_settings()
completions_settings.msw_roughness = 12.34
completions_settings.msw_liner_diameter = 0.2222
completions_settings.well_name_for_export = "file name"
completions_settings.group_name_for_export = "msj"
completions_settings.well_type_for_export = "GAS"
completions_settings.update()  # Commit updates back to ResInsight

# Optionally update the MSW settings
msw_settings = well_path.msw_settings()
msw_settings.custom_values_for_lateral = False
msw_settings.enforce_max_segment_length = False
msw_settings.liner_diameter = 0.152
msw_settings.max_segment_length = 200
msw_settings.pressure_drop = "HF-"
msw_settings.reference_md_type = "GridEntryPoint"
msw_settings.roughness_factor = 1e-05
msw_settings.user_defined_reference_md = 0
msw_settings.update()

# export completions
cases = resinsight.project.cases()

for case in cases:
    print("Case name: ", case.name)
    print("Case id: ", case.id)

    case.export_well_path_completions(
        time_step=0,
        well_path_names=["Test Well-1 Y1"],
        file_split="UNIFIED_FILE",
        include_perforations=True,
        # Replace the following with a valid path
        custom_file_name="f:/scratch/2023-11-02/myfile.myext",
    )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import SimulationWell_pb2 as SimulationWell__pb2


class SimulationWellStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetSimulationWellStatus = channel.unary_unary(
                '/rips.SimulationWell/GetSimulationWellStatus',
                request_serializer=SimulationWell__pb2.SimulationWellRequest.SerializeToString,
                response_deserializer=SimulationWell__pb2.SimulationWellStatus.FromString,
                )
        self.GetSimulationWellCells = channel.unary_unary(
                '/rips.SimulationWell/GetSimulationWellCells',
                request_serializer=SimulationWell__pb2.SimulationWellRequest.SerializeToString,
                response_deserializer=SimulationWell__pb2.SimulationWellCellInfoArray.FromString,
                )


class SimulationWellServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetSimulationWellStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSimulationWellCells(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SimulationWellServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetSimulationWellStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSimulationWellStatus,
                    request_deserializer=SimulationWell__pb2.SimulationWellRequest.FromString,
                    response_serializer=SimulationWell__pb2.SimulationWellStatus.SerializeToString,
            ),
            'GetSimulationWellCells': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSimulationWellCells,
                    request_deserializer=SimulationWell__pb2.SimulationWellRequest.FromString,
                    response_serializer=SimulationWell__pb2.SimulationWellCellInfoArray.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'rips.SimulationWell', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SimulationWell(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetSimulationWellStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/rips.SimulationWell/GetSimulationWellStatus',
            SimulationWell__pb2.SimulationWellRequest.SerializeToString,
            SimulationWell__pb2.SimulationWellStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSimulationWellCells(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/rips.SimulationWell/GetSimulationWellCells',
            SimulationWell__pb2.SimulationWellRequest.SerializeToString,
            SimulationWell__pb2.SimulationWellCellInfoArray.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Load ResInsight Processing Server Client Library
import rips
import time

# Connect to ResInsight instance
resinsight = rips.Instance.find()

# Create a modeled well path and add well path targets
# The coordinates are based on the Norne case
# Add a lateral to the main well path

well_path_coll = resinsight.project.descendants(rips.WellPathCollection)[0]

editable_well_paths = well_path_coll.descendants(rips.ModeledWellPath)
if editable_well_paths:
    # Use the first well path if it exists
    well_path = editable_well_paths[0]
else:
    # Create a well path
    well_path = well_path_coll.add_new_object(rips.ModeledWellPath)
    well_path.name = "Test Well-1"
    well_path.update()

    geometry = well_path.well_path_geometry()

    reference_point = geometry.reference_point
    reference_point[0] = 457196
    reference_point[1] = 7322270
    reference_point[2] = 2742
    geometry.update()

    coord = [0, 0, 0]
    geometry.append_well_target(coord)

    coord = [454.28, 250, -10]
    target = geometry.append_well_target(coord)

    coord = [1054.28, 250, -50]
    target = geometry.append_well_target(coord)


# Add a fishbone completion to the well path with default settings
well_path.append_fishbones([4000.0, 4100.0, 4200.0])

# Optional settings on the collection level that will affect all installed fishbones
fishbones_collection = well_path.completions().fishbones()
fishbones_collection.main_bore_skin_factor = 0.1
fishbones_collection.update()

# Drilling type is one of [STANDARD, EXTENDED, ACID_JETTING]
drilling_type = "ACID_JETTING"

sub_locations = [3500.0, 3550.0, 3600.0, 3700.0]
fishbones = fishbones_collection.append_fishbones(sub_locations, drilling_type)

fishbones.lateral_diameter = 47.0
fishbones.lateral_skin_factor = 27.0
fishbones.update()

# Optionally set the fixed start location for all fishbones
fishbones_collection.set_fixed_start_location(2900.0)

# Export completions
cases = resinsight.project.cases()

for case in cases:
    print("Case name: ", case.name)
    print("Case id: ", case.id)

    case.export_well_path_completions(
        time_step=0,
        well_path_names=["Test Well-1 Y1"],
        file_split="UNIFIED_FILE",
        include_perforations=True,
        custom_file_name="f:/scratch/2023-11-02/myfile.myext",
    )



# pylint: disable=no-self-use
"""
The main entry point for ResInsight connections
The Instance class contained have static methods launch and find for
creating connections to ResInsight
"""

from __future__ import annotations
import os
import socket
import logging
import time
import tempfile
import signal
import sys
import json

import grpc

import App_pb2
import App_pb2_grpc
import Commands_pb2
import Commands_pb2_grpc
from Definitions_pb2 import Empty

import RiaVersionInfo

from .project import Project
from .retry_policy import ExponentialBackoffRetryPolicy
from .grpc_retry_interceptor import RetryOnRpcErrorClientInterceptor
from .generated.generated_classes import CommandRouter

from typing import List, Optional, Tuple
from typing_extensions import Self
from pathlib import Path


class Instance:
    """The ResInsight Instance class. Use to launch or find existing ResInsight instances

    Attributes:
        launched (bool): Tells us whether the application was launched as a new process.
            If the application was launched we may need to close it when exiting the script.
        commands (Commands): Command executor. Set when creating an instance.
        project (Project): Current project in ResInsight.
            Set when creating an instance and updated when opening/closing projects.
    """

    @staticmethod
    def __is_port_in_use(port: int) -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as my_socket:
            my_socket.settimeout(0.2)
            return my_socket.connect_ex(("localhost", port)) == 0

    @staticmethod
    def __is_valid_port(port: int) -> bool:
        location = "localhost:" + str(port)
        channel = grpc.insecure_channel(
            location, options=[("grpc.enable_http_proxy", False)]
        )
        app = App_pb2_grpc.AppStub(channel)
        try:
            app.GetVersion(Empty(), timeout=1)
        except grpc.RpcError:
            return False
        return True

    @staticmethod
    def __read_port_number_from_file(file_path: str, max_attempts: int) -> int:
        retry_count = 0
        while not os.path.exists(file_path) and retry_count < max_attempts:
            time.sleep(1)
            retry_count = retry_count + 1

        if os.path.isfile(file_path):
            with open(file_path) as f:
                value = f.readline()
                return int(value)

        if retry_count == max_attempts:
            print(
                "Waiting for port number file timed out after {} seconds. File: {}".format(
                    max_attempts, file_path
                )
            )

        return -1

    @staticmethod
    def __kill_process(pid: int) -> None:
        """
        Kill the process with a given pid.
        """
        if hasattr(signal, "CTRL_C_EVENT"):
            # windows does not have kill
            os.kill(pid, signal.CTRL_C_EVENT)
        else:
            # linux/unix
            os.kill(pid, signal.SIGKILL)

    @staticmethod
    def launch(
        resinsight_executable: str = "",
        console: bool = False,
        launch_port: int = 0,
        init_timeout: int = 300,
        command_line_parameters: List[str] = [],
    ) -> Optional[Instance]:
        """Launch a new Instance of ResInsight. This requires the environment variable
        RESINSIGHT_EXECUTABLE to be set or the parameter resinsight_executable to be provided.
        The RESINSIGHT_GRPC_PORT environment variable can be set to an alternative port number.

        Args:
            resinsight_executable (str): Path to a valid ResInsight executable. If set
                will take precedence over what is provided in the RESINSIGHT_EXECUTABLE
                environment variable.
            console (bool): If True, launch as console application, without GUI.
            launch_port(int): If 0, GRPC will find an available port.
                             If -1, use the default port 50051 or RESINSIGHT_GRPC_PORT
                             If anything else, ResInsight will try to launch with the specified portnumber.
            init_timeout: Number of seconds to wait for initialization before timing out.
            command_line_parameters(list): Additional parameters as string entries in the list.
        Returns:
            Instance: an instance object if it worked. None if not.
        """

        requested_port: int = 50051
        port_env = os.environ.get("RESINSIGHT_GRPC_PORT")
        if port_env:
            requested_port = int(port_env)
        if launch_port != -1:
            requested_port = launch_port

        if not resinsight_executable:
            filename = Path(sys.prefix) / "share" / "rips" / "rips_config.json"
            if filename.is_file():
                f = open(filename)
                data = json.load(f)
                resinsight_executable = data["resinsight_executable"]
                if resinsight_executable:
                    print(
                        "In './share/rips/rips_config.json', found resinsight_executable:",
                        resinsight_executable,
                    )

        if not resinsight_executable:
            resinsight_executable_from_env = os.environ.get("RESINSIGHT_EXECUTABLE")
            if not resinsight_executable_from_env:
                print(
                    "ERROR: Could not launch ResInsight because the environment variable"
                    " RESINSIGHT_EXECUTABLE is not set"
                )
                return None
            else:
                resinsight_executable = resinsight_executable_from_env

        print("Trying to launch", resinsight_executable)
        with tempfile.TemporaryDirectory() as tmp_dir_path:
            port_number_file = tmp_dir_path + "/portnumber.txt"
            parameters: List[str] = [
                "ResInsight",
                "--server",
                str(requested_port),
                "--portnumberfile",
                str(port_number_file),
            ] + command_line_parameters
            if console:
                print("Launching as console app")
                parameters.append("--console")

            # Stringify all parameters
            for i in range(0, len(parameters)):
                parameters[i] = str(parameters[i])

            pid = os.spawnv(os.P_NOWAIT, resinsight_executable, parameters)
            if pid:
                port = Instance.__read_port_number_from_file(
                    port_number_file, init_timeout
                )
                if port == -1:
                    print("Unable to read port number. Launch failed.")
                    # Need to kill the process using PID since there is no  GRPC connection to use.
                    Instance.__kill_process(pid)
                else:
                    instance = Instance(port=port, launched=True)
                    return instance
        return None

    @staticmethod
    def find(start_port: int = 50051, end_port: int = 50071) -> Optional[Instance]:
        """Search for an existing Instance of ResInsight by testing ports.

        By default we search from port 50051 to 50071 or if the environment
        variable RESINSIGHT_GRPC_PORT is set we search
        RESINSIGHT_GRPC_PORT to RESINSIGHT_GRPC_PORT+20

        Args:
            start_port (int): start searching from this port
            end_port (int): search up to but not including this port
        """
        port_env = os.environ.get("RESINSIGHT_GRPC_PORT")
        if port_env:
            print("Got port " + port_env + " from environment")
            start_port = int(port_env)
            end_port = start_port + 20

        for try_port in range(start_port, end_port):
            print("Trying port " + str(try_port))
            if Instance.__is_port_in_use(try_port) and Instance.__is_valid_port(
                try_port
            ):
                return Instance(port=try_port)

        print(
            "Error: Could not find any ResInsight instances responding between ports "
            + str(start_port)
            + " and "
            + str(end_port)
        )
        return None

    def __execute_command(self, **command_params):
        return self.commands.Execute(Commands_pb2.CommandParams(**command_params))

    def __check_version(self) -> Tuple[bool, bool]:
        try:
            major_version_ok = self.major_version() == int(
                RiaVersionInfo.RESINSIGHT_MAJOR_VERSION
            )
            minor_version_ok = self.minor_version() == int(
                RiaVersionInfo.RESINSIGHT_MINOR_VERSION
            )
            return True, major_version_ok and minor_version_ok
        except grpc.RpcError:
            return False, False

    def __init__(self, port: int = 50051, launched: bool = False) -> None:
        """Attempts to connect to ResInsight at a specific port on localhost

        Args:
            port(int): port number
        """
        logging.basicConfig()
        self.location: str = "localhost:" + str(port)

        self.channel = grpc.insecure_channel(
            self.location, options=[("grpc.enable_http_proxy", False)]
        )
        self.launched = launched
        self.commands = Commands_pb2_grpc.CommandsStub(self.channel)

        # Main version check package
        self.app = App_pb2_grpc.AppStub(self.channel)

        self._check_connection_and_version(self.channel, launched, self.location)

        # Intercept UNAVAILABLE errors and retry on failures
        interceptors = (
            RetryOnRpcErrorClientInterceptor(
                retry_policy=ExponentialBackoffRetryPolicy(
                    min_backoff=100, max_backoff=5000, max_num_retries=20
                ),
                status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
            ),
        )

        intercepted_channel = grpc.intercept_channel(self.channel, *interceptors)

        # Recreate command stubs with the retry policy
        self.commands = Commands_pb2_grpc.CommandsStub(intercepted_channel)

        # Service packages
        self.project = Project.create(intercepted_channel)

        # Command Router object used as entry point for independent processing functions
        self.command_router = CommandRouter(
            self.app.GetPdmObject(Empty()), intercepted_channel
        )

        path = os.getcwd()
        self.set_start_dir(path=path)

    def _check_connection_and_version(
        self, channel: grpc.Channel, launched: bool, location: str
    ) -> None:
        connection_ok = False
        version_ok = False

        retry_policy = ExponentialBackoffRetryPolicy()
        if self.launched:
            for num_tries in range(0, retry_policy.num_retries()):
                connection_ok, version_ok = self.__check_version()
                if connection_ok:
                    break
                retry_policy.sleep(num_tries)
        else:
            connection_ok, version_ok = self.__check_version()

        if not connection_ok:
            if self.launched:
                raise Exception(
                    "Error: Could not connect to resinsight at ",
                    location,
                    ".",
                    retry_policy.time_out_message(),
                )
            raise Exception("Error: Could not connect to resinsight at ", location)
        if not version_ok:
            raise Exception(
                "Error: Wrong Version of ResInsight at ",
                location,
                "Executable : " + self.version_string(),
                " ",
                "rips : " + self.client_version_string(),
            )

    def __version_message(self) -> App_pb2.Version:
        return self.app.GetVersion(Empty())

    def set_start_dir(self, path: str):
        """Set current start directory

        Arguments:
            path (str): path to directory

        """
        return self.__execute_command(
            setStartDir=Commands_pb2.FilePathRequest(path=path)
        )

    def set_export_folder(
        self, export_type: str, path: str, create_folder: bool = False
    ):
        """
        Set the export folder used for all export functions

        **Parameters**::

            Parameter        | Description                                  | Type
            ---------------- | -------------------------------------------- | -----
            export_type      | String specifying what to export             | String
            path             | Path to folder                               | String
            create_folder    | Create folder if it doesn't exist?           | Boolean

        **Enum export_type**::

            Option          | Description
            --------------- | ------------
            "COMPLETIONS"   |
            "SNAPSHOTS"     |
            "PROPERTIES"    |
            "STATISTICS"    |

        """
        return self.__execute_command(
            setExportFolder=Commands_pb2.SetExportFolderRequest(
                type=export_type, path=path, createFolder=create_folder
            )
        )

    def set_main_window_size(self, width: int, height: int):
        """
        Set the main window size in pixels

        **Parameters**::

            Parameter | Description      | Type
            --------- | ---------------- | -----
            width     | Width in pixels  | Integer
            height    | Height in pixels | Integer

        """
        return self.__execute_command(
            setMainWindowSize=Commands_pb2.SetWindowSizeParams(
                width=width, height=height
            )
        )

    def set_plot_window_size(self, width: int, height: int):
        """
        Set the plot window size in pixels

        **Parameters**::

            Parameter | Description      | Type
            --------- | ---------------- | -----
            width     | Width in pixels  | Integer
            height    | Height in pixels | Integer
        """
        return self.__execute_command(
            setPlotWindowSize=Commands_pb2.SetWindowSizeParams(
                width=width, height=height
            )
        )

    def major_version(self) -> int:
        """Get an integer with the major version number"""
        return int(self.__version_message().major_version)

    def minor_version(self) -> int:
        """Get an integer with the minor version number"""
        return int(self.__version_message().minor_version)

    def patch_version(self) -> int:
        """Get an integer with the patch version number"""
        return int(self.__version_message().patch_version)

    def version_string(self) -> str:
        """Get a full version string, i.e. 2019.04.01"""
        return (
            str(self.major_version())
            + "."
            + str(self.minor_version())
            + "."
            + str(self.patch_version())
        )

    def client_version_string(self) -> str:
        """Get a full version string, i.e. 2019.04.01"""
        version_string: str = RiaVersionInfo.RESINSIGHT_MAJOR_VERSION + "."
        version_string += RiaVersionInfo.RESINSIGHT_MINOR_VERSION + "."
        version_string += RiaVersionInfo.RESINSIGHT_PATCH_VERSION
        return version_string

    def exit(self):
        """Tell ResInsight instance to quit"""
        print("Telling ResInsight to Exit")
        return self.app.Exit(Empty())

    def is_console(self) -> bool:
        """Returns true if the connected ResInsight instance is a console app"""
        return bool(
            self.app.GetRuntimeInfo(Empty()).app_type
            == App_pb2.ApplicationTypeEnum.Value("CONSOLE_APPLICATION")
        )

    def is_gui(self) -> bool:
        """Returns true if the connected ResInsight instance is a GUI app"""
        return bool(
            self.app.GetRuntimeInfo(Empty()).app_type
            == App_pb2.ApplicationTypeEnum.Value("GUI_APPLICATION")
        )
